import os
from io import StringIO
from unittest import TestCase
from unittest.mock import patch

from shub_workflow.base import WorkFlowManager

from .utils.contexts import script_args


@patch("shub_workflow.base.WorkFlowManager.get_job_tags")
@patch("shub_workflow.base.WorkFlowManager._update_metadata")
class WorkFlowManagerTest(TestCase):
    def setUp(self):
        os.environ["SH_APIKEY"] = "ffff"
        os.environ["PROJECT_ID"] = "999"

    @patch("sys.stderr", new_callable=StringIO)
    def test_name_required_not_set(self, mocked_stderr, mocked_update_metadata, mocked_get_job_tags):
        class TestManager(WorkFlowManager):
            def workflow_loop(self):
                return True

        mocked_get_job_tags.side_effect = [[], []]

        with script_args([]):
            with self.assertRaises(SystemExit):
                TestManager()
        self.assertTrue("the following arguments are required: name" in mocked_stderr.getvalue())

    def test_name_required_set(self, mocked_update_metadata, mocked_get_job_tags):
        class TestManager(WorkFlowManager):
            def workflow_loop(self):
                return True

        mocked_get_job_tags.side_effect = [[], []]

        with script_args(["my_fantasy_name"]):
            manager = TestManager()
        self.assertEqual(manager.name, "my_fantasy_name")



#!/usr/bin/env python
"""
Generate deliverables

Usage:

    deliver.py <spider> [<spider> ... ][options]

For listing valid values for spider argument, just run deliver.py without arguments.

This script does the following:

* searches for untagged finished jobs that corresponds to the given spiders and same flow id as inherited or passed
  to the delivery job.
* gets all their items generated and uploads to customer s3 bucket, gs buckets or local folder, in files of given max
  size (either in items or bytes)
* tags processed jobs with 'delivered' tag, so they are not processed again (tag can be customized)

Deliver class is meant to be subclasses for overriding default behaviors, either via overriding of configuration class
attributes or instance methods.

configuration class attributes
------------------------------

DeliverScript.output_prefix - Target output prefix. Supports s3://..., gs:// and local absolute or relative folders
                              (required)
DeliverScript.success_file - A boolean. Whether or not to generate a finall _SUCCESS file
                             (optional, False by default)

Following environment variables are also required:

In case of using s3: AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, SH_APIKEY
In case of using gs: GOOGLE_APPLICATION_CREDENTIALS (you can use helper method
shub_workflow.deliver.gcstorage.set_credential_file_environ)

The default file format are gzipped json list. The default file name format will be:

<output prefix>/<spider name>/<formatted timestamp>/<file number>.jsonl.gz

The formatted timestamp corresponds to the time where the delivery script starts to run.

These defaults are only there for providing a fast deployable delivery script. But mosts project has its own
requirements, so the DeliverScript is designed for easy overriding of methods for very flexible customization.


"""
import json
import sys
import os
import gzip
import datetime
import logging
from tempfile import mktemp

from shub_workflow.deliver.futils import mv_file, touch
from shub_workflow.deliver.dupefilter import SqliteDictDupesFilter
from shub_workflow.deliver.base import BaseDeliverScript


_LOG = logging.getLogger(__name__)
_LOG.setLevel(logging.INFO)


class OutputFile:

    max_filesize_items = 1000000
    max_filesize_bytes = 1000000000
    mv_file_kwargs = None
    gzipped_output = True

    def __init__(self, output_prefix, keyprefix, testmode=False):
        self.__output_prefix = output_prefix
        self.__key_prefix = keyprefix

        self.__filename = mktemp()
        self.__file = self.create_new_file()
        self.__filecount = 0
        self.__items_count = 0
        self.__size_bytes = 0
        self.__testmode = testmode

        spath = self.__key_prefix.split("/")[:-1]
        spath.append("_SUCCESS")
        self.success_file = os.path.join(*spath)

    @property
    def items_count(self):
        return self.__items_count

    @property
    def filename(self):
        return self.__filename

    @property
    def key_prefix(self):
        return self.__key_prefix

    @property
    def filecount(self):
        return self.__filecount

    def create_new_file(self):
        if self.gzipped_output:
            return gzip.open(self.filename, "wb")
        return open(self.filename, "w")

    def write(self, item):
        line = json.dumps(item) + "\n"
        if self.gzipped_output:
            self.__file.write(line.encode("utf8"))
        else:
            self.__file.write(line)
        self.__items_count += 1
        self.__size_bytes += sys.getsizeof(line)
        if (
            self.max_filesize_items > 0
            and self.__items_count == self.max_filesize_items
            or self.max_filesize_bytes > 0
            and self.__size_bytes >= self.max_filesize_bytes
        ):
            self.flush(new_file=True)

    def flush(self, new_file=False):
        if self.__items_count > 0:
            self.__file.close()
            self._upload_file()
            self.__filecount += 1
            self.__items_count = 0
            self.__size_bytes = 0
            if new_file:
                self.__file = self.create_new_file()
        elif not new_file:
            self.__file.close()

    def gen_keyname(self):
        return self.key_prefix + "%05d.jsonl.gz" % self.filecount

    def _upload_file(self):
        keyname = self.gen_keyname()
        destination = os.path.join(self.__output_prefix, keyname)
        if not self.__testmode:
            kwargs = self.mv_file_kwargs or {}
            mv_file(self.filename, destination, **kwargs)
        _LOG.info(f"Saved {self.__items_count} items in {destination}")


class OutputFileDict:

    outputfile_class = OutputFile

    def __init__(self, output_prefix, testmode=False):
        self.__output_prefix = output_prefix
        self.__testmode = testmode
        self.__outputfiles = {}

    def __getitem__(self, key):
        if key not in self.__outputfiles:
            self.__outputfiles[key] = self.outputfile_class(
                output_prefix=self.__output_prefix, keyprefix=key, testmode=self.__testmode
            )
        return self.__outputfiles[key]

    def values(self):
        return self.__outputfiles.values()

    def keys(self):
        return self.__outputfiles.keys()

    def pop(self, key):
        return self.__outputfiles.pop(key)


class DeliverScript(BaseDeliverScript):

    write_success_file = False
    output_prefix = ""
    outputfiledict_class = OutputFileDict

    default_sh_chunk_size = 1_000

    def __init__(self):
        super().__init__()

        self.output_files = self.set_output_files()
        self.__start_datetime = datetime.datetime.now()

        self.itemcount = 0
        self.filecount = 0
        self.dupes_filter = {i: SqliteDictDupesFilter() for i in self.args.filter_dupes_by_field}
        self.success_files = []

    @property
    def description(self):
        return __doc__

    @property
    def start_datetime(self):
        return self.__start_datetime

    def set_output_files(self):
        return self.outputfiledict_class(self.args.output_prefix, self.args.test_mode)

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument("--output-prefix", help="Delivery prefix.", default=self.output_prefix)
        self.argparser.add_argument(
            "--filter-dupes-by-field",
            default=[],
            help="Dedupe by any of the given item field. Can be given multiple times",
        )
        self.argparser.add_argument("--one-file-per-job", action="store_true", help="Generate one file per job.")
        self.argparser.add_argument(
            "--sh-chunk-size",
            type=int,
            default=self.default_sh_chunk_size,
            help=(
                "Chunk/page size for downloading items from Scrapy Cloud. For tweaking memory consumption and speed."
                " Note that the performance will depend on the sizes of individual items in the cloud."
            ),
        )

    def gen_keyprefix(self, scrapername, job, item):
        formatted_datetime = self.start_datetime.strftime("%Y-%m-%dT%H:%M:%S")
        return os.path.join(scrapername, formatted_datetime)

    def process_job_items(self, scrapername, spider_job):
        first_keyprefix = None
        job_item_count = 0
        chunks = spider_job.items.list_iter(chunksize=self.args.sh_chunk_size)
        items_iter = (item for chunk in chunks for item in chunk)
        for item in items_iter:
            seen = False
            for field in self.dupes_filter.keys():
                if field in item:
                    if item[field] in self.dupes_filter[field]:
                        seen = True
                    else:
                        self.dupes_filter[field].add(item[field])
            if seen:
                continue

            try:
                keyprefix = self.gen_keyprefix(scrapername, spider_job, item)
                first_keyprefix = first_keyprefix or keyprefix
            except KeyError as e:
                if first_keyprefix is None:
                    _LOG.info("Skipped job: %s. %s", spider_job.key, str(e))
                    return
                keyprefix = first_keyprefix
            self.output_files[keyprefix].write(item)
            self.itemcount += 1
            job_item_count += 1
            if self.itemcount % 100000 == 0:
                _LOG.info("Processed %d items.", self.itemcount)
        _LOG.info("Processed all %d items of spider job %s", job_item_count, spider_job.key)

    def process_spider_jobs(self, scrapername):
        super().process_spider_jobs(scrapername)
        _LOG.info("Total Processed items for spider %s: %d", scrapername, self.itemcount)
        for ofile in self.output_files.values():
            ofile.flush()
            if self.write_success_file:
                self.success_files.add(ofile.success_file)

    def on_close(self):
        for success_file in self.success_files:
            remote_success_file = os.path.join(self.args.output_prefix, success_file)
            if not self.args.test_mode:
                touch(remote_success_file)
            _LOG.info(f"Created {remote_success_file}")
        super().on_close()


# for compatibility with older versions. New DeliverScript handles indistictly local files, s3 and gs.
class S3DeliverScript(DeliverScript):

    s3_bucket_name = None

    def __init__(self):
        self.output_prefix = self.s3_bucket_name
        super().__init__()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format=("%(asctime)s [%(levelname)s] %(pathname)s:%(lineno)d %(message)s"))
    deliver = DeliverScript()
    deliver.run()



import json
import logging
from uuid import uuid1
from typing import Set

from collection_scanner import CollectionScanner
from scrapinghub.client.exceptions import NotFound

from shub_workflow.script import BaseScript
from shub_workflow.deliver.dupefilter import SqliteDictDupesFilter

_LOG = logging.getLogger(__name__)
_LOG.setLevel(logging.INFO)


class BaseDeliverScript(BaseScript):

    DELIVERED_TAG = "delivered"
    SCRAPERNAME_NARGS = "+"

    # print log every given items processed
    LOG_EVERY = 1000

    # define here the fields used to deduplicate items. All them compose the dedupe key.
    # target item values must be strings.
    # for changing behavior, override is_seen_item()
    DEDUPE_KEY_BY_FIELDS = ()

    MAX_PROCESSED_ITEMS = float("inf")

    SEEN_ITEMS_CLASS = SqliteDictDupesFilter

    def __init__(self):
        super().__init__()
        self._all_jobs_to_tag = []
        self.total_items_count = 0
        self.total_dupe_filtered_items_count = 0
        self.seen_items = self.SEEN_ITEMS_CLASS()

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument("scrapername", help="Target scraper names", nargs=self.SCRAPERNAME_NARGS)
        self.argparser.add_argument(
            "--test-mode",
            action="store_true",
            help="Run in test mode (performs all processes, but doesn't upload files nor consumes jobs)",
        )

    def get_target_tags(self):
        return []

    def get_delivery_spider_jobs(self, scrapername, target_tags):
        if self.flow_id:
            flow_id_tag = [f"FLOW_ID={self.flow_id}"]
            target_tags = flow_id_tag + target_tags
        yield from self.get_jobs_with_tags(scrapername, target_tags, state=["finished"], lacks_tag=[self.DELIVERED_TAG])

    def process_spider_jobs(self, scrapername):
        target_tags = self.get_target_tags()
        for sj in self.get_delivery_spider_jobs(scrapername, target_tags):
            self.process_job_items(scrapername, sj)
            if not self.args.test_mode:
                self._all_jobs_to_tag.append(sj.key)
            if self.total_items_count >= self.MAX_PROCESSED_ITEMS:
                break

    def get_item_unique_key(self, item):
        key = tuple(item[f] for f in self.DEDUPE_KEY_BY_FIELDS)
        return ",".join(key)

    def is_seen_item(self, item):
        key = self.get_item_unique_key(item)
        return key and key in self.seen_items

    def add_seen_item(self, item):
        key = self.get_item_unique_key(item)
        if key:
            self.seen_items.add(key)

    def process_job_items(self, scrapername, spider_job):
        for item in spider_job.items.iter():
            if self.is_seen_item(item):
                self.total_dupe_filtered_items_count += 1
            else:
                self.on_item(item, scrapername)
                self.add_seen_item(item)
            self.total_items_count += 1
            if self.total_items_count % self.LOG_EVERY == 0:
                _LOG.info(f"Processed {self.total_items_count} items.")

    def on_item(self, item, scrapername):
        print(json.dumps(item))

    def run(self):
        for scrapername in self.args.scrapername:
            _LOG.info(f"Processing spider {scrapername}")
            self.process_spider_jobs(scrapername)
        self.on_close()

    def on_close(self):
        _LOG.info(f"Processed a total of {self.total_items_count} items.")
        if self.DEDUPE_KEY_BY_FIELDS:
            _LOG.info(f"A total of {self.total_dupe_filtered_items_count} items were duplicated.")
        jcount = 0
        for jkey in self._all_jobs_to_tag:
            self.add_job_tags(jkey, tags=[self.DELIVERED_TAG])
            jcount += 1
            if jcount % 100 == 0:
                _LOG.info("Marked %d jobs as delivered", jcount)
        if hasattr(self.seen_items, "close"):
            self.seen_items.close()


class CachedDeliveredTagsMixin:

    """
    Usage:

    class MyDeliveryScript(CachedDeliveredTagsMixin, BaseDeliverScript):
        pass

    Under situations where you need speed and you have lots of jobs to deliver from,
    this mixin caches in a collection the delivered jobs instead of tagging them
    one by one.

    An external job scheduled from time to time may take care of tagging jobs while
    cleaning the collection. This is the recommended approach. Use the class
    SyncDeliveredScript for creating a script for that purpose.
    """

    TAGS_COLLECTION = "delivered"

    def __init__(self):
        super().__init__()
        self._cache = None
        self._to_store: Set = set()

    def get_delivered_cached(self):
        self._cache = set()
        try:
            for batch in CollectionScanner(
                collection_name=self.TAGS_COLLECTION,
                project_id=super().get_project().key,
            ).scan_collection_batches():
                for record in batch:
                    self._cache.update(record["data"])
        except NotFound:
            pass
        _LOG.info(f"Loaded {len(self._cache)} keys from delivered cache collection.")

    def add_job_tags(self, jobkey=None, tags=None):
        if jobkey is None or tags != [self.DELIVERED_TAG]:
            super().add_job_tags(jobkey, tags)
        else:
            if self._cache is None:
                self.get_delivered_cached()
            self._cache.add(jobkey)
            self._to_store.add(jobkey)

    def sync(self):
        if self._to_store:
            col = super().get_project().collections.get_store(self.TAGS_COLLECTION)
            to_store_all = list(self._to_store)
            self._to_store = set()
            to_store_count = len(to_store_all)
            while to_store_all:
                to_store, to_store_all = to_store_all[:100], to_store_all[100:]
                key = str(uuid1())
                record = {"_key": key, "data": list(to_store)}
                col.set(record)
            _LOG.info(f"Synced delivered cache ({to_store_count} jobs).")

    def get_delivery_spider_jobs(self, scrapername, target_tags):
        if self._cache is None:
            self.get_delivered_cached()
        for sj in super().get_delivery_spider_jobs(scrapername, target_tags):
            if sj.key not in self._cache:
                yield sj

    def on_close(self):
        super().on_close()
        self.sync()


class SyncDeliveredScript(BaseScript):

    TAGS_COLLECTION = "delivered"
    DELIVERED_TAG = "delivered"

    def run(self):
        store = self.get_project().collections.get_store(self.TAGS_COLLECTION)
        try:
            for batch in CollectionScanner(
                collection_name=self.TAGS_COLLECTION,
                project_id=super().get_project().key,
                meta=["_key"],
            ).scan_collection_batches():
                for record in batch:
                    for key in record["data"]:
                        self.add_job_tags(key, [self.DELIVERED_TAG])
                    store.delete([record["_key"]])
                    _LOG.info(f"Synced {len(record['data'])} jobs from {record['_key']}.")
        except NotFound:
            pass


if __name__ == "__main__":
    logging.basicConfig(format="%(asctime)s %(name)s [%(levelname)s]: %(message)s", level=logging.INFO)
    script = SyncDeliveredScript()
    script.run()



"""
Base script class for spiders crawl managers.
"""
import abc
import json
import logging

from bloom_filter import BloomFilter

from shub_workflow.base import WorkFlowManager
from shub_workflow.utils import hashstr


_LOG = logging.getLogger(__name__)
_LOG.setLevel(logging.INFO)


def get_spider_args_from_params(params):
    spider_args = params.copy()
    for key in "spider", "units", "project_id", "tags":
        spider_args.pop(key, None)
    return spider_args


def get_jobseq(tags):
    jobseq, repetition = 0, 0
    for tag in tags:
        if tag.startswith("JOBSEQ="):
            tag = tag.replace("JOBSEQ=", "")
            jobseq, *rep = tag.split(".r")
            rep.append(0)
            repetition = int(rep[0])
            jobseq = int(jobseq)
    return jobseq, repetition


class CrawlManager(WorkFlowManager):
    """
    Schedules a single spider job. If loop mode is enabled, it will shutdown only after the scheduled spider
    finished. Close reason of the manager will be inherited from spider one.
    """

    spider = None

    def __init__(self):
        super().__init__()
        self.__close_reason = None

        # running jobs represents the state of a crawl manager.
        # a dict job key : (spider, spider_args_override)
        self._running_job_keys = {}

    @property
    def description(self):
        return self.__doc__

    def add_argparser_options(self):
        super().add_argparser_options()
        if self.spider is None:
            self.argparser.add_argument("spider", help="Spider name")
        self.argparser.add_argument("--spider-args", help="Spider arguments dict in json format", default="{}")
        self.argparser.add_argument("--job-settings", help="Job settings dict in json format", default="{}")
        self.argparser.add_argument("--units", help="Set default number of ScrapyCloud units for each job", type=int)

    def parse_args(self):
        args = super().parse_args()
        if self.spider is None:
            self.spider = args.spider
        return args

    def get_spider_args(self, override=None):
        spider_args = json.loads(self.args.spider_args)
        if override:
            spider_args.update(override)
        return spider_args

    def get_job_settings(self, override=None):
        job_settings = json.loads(self.args.job_settings)
        if override:
            job_settings.update(override)
        return job_settings

    def schedule_spider(self, spider=None, spider_args_override=None):
        if spider_args_override is not None:
            spider_args_override.copy()
        spider = spider or self.spider
        spider_args = self.get_spider_args(spider_args_override)
        job_settings_override = spider_args.pop("job_settings", None)
        job_settings = self.get_job_settings(job_settings_override)
        spider_args["job_settings"] = job_settings
        spider_args["units"] = spider_args.get("units", self.args.units)
        jobkey = super().schedule_spider(spider, **spider_args)
        if jobkey is not None:
            self._running_job_keys[jobkey] = spider, spider_args_override
        return jobkey

    def check_running_jobs(self):
        outcomes = {}
        running_job_keys = list(self._running_job_keys)
        while running_job_keys:
            jobkey = running_job_keys.pop()
            if (outcome := self.is_finished(jobkey)) is not None:
                spider, spider_args_override = self._running_job_keys.pop(jobkey)
                if outcome in self.failed_outcomes:
                    _LOG.warning(f"Job {jobkey} finished with outcome {outcome}.")
                    if spider_args_override is not None:
                        spider_args_override = spider_args_override.copy()
                    self.bad_outcome_hook(spider, outcome, spider_args_override, jobkey)
                outcomes[jobkey] = outcome
        _LOG.info(f"There are {len(self._running_job_keys)} jobs still running.")

        return outcomes

    def bad_outcome_hook(self, spider, outcome, spider_args_override, jobkey):
        if self.__close_reason is None:
            self.__close_reason = outcome

    def workflow_loop(self):
        outcomes = self.check_running_jobs()
        if outcomes:
            return False
        if not self._running_job_keys:
            self.schedule_spider()
        return True

    def resume_workflow(self):
        running_jobs = []
        count = 0
        for job in self.get_owned_jobs(state=["running", "pending"], meta=["spider_args", "tags", "spider"]):
            key = job["key"]
            spider_args_override = job.get("spider_args", {}).copy()
            spider_args_override["tags"] = job["tags"]
            self._running_job_keys[key] = job["spider"], spider_args_override
            _LOG.info(f"added running job {key}")
            running_jobs.append(job)
            count += 1
        _LOG.info(f"Added a total of {count} running jobs.")
        return running_jobs

    def on_close(self):
        job = self.get_job()
        if job:
            close_reason = self.__close_reason
            self.finish(close_reason=close_reason)


class PeriodicCrawlManager(CrawlManager):
    """
    Schedule a spider periodically, waiting for the previous job to finish before scheduling it again with same
    parameters. Don't forget to set loop mode.
    """

    def bad_outcome_hook(self, spider, outcome, spider_args_override, jobkey):
        pass

    def workflow_loop(self):
        self.check_running_jobs()
        if not self._running_job_keys:
            self.schedule_spider()
        return True

    def on_close(self):
        pass


class GeneratorCrawlManager(CrawlManager):
    """
    Schedule a spider periodically, each time with different parameters yielded by a generator, until stop.
    Number of simultaneos spider jobs will be limited by max running jobs (see WorkFlowManager).
    Instructions:
    - Override set_parameters_gen() method. It must implement a generator of dictionaries, each one being
      the spider arguments (argument name: argument value) passed to the spider on each successive job.
    - Don't forget to set loop mode and max jobs (which defaults to infinite).
    """

    MAX_TOTAL_JOBS = 1000000

    def __init__(self):
        super().__init__()
        self.__parameters_gen = self.set_parameters_gen()
        self.__additional_jobs = []
        self.__next_job_seq = 1
        self.__jobids = BloomFilter(max_elements=self.MAX_TOTAL_JOBS, error_rate=0.001)

    def bad_outcome_hook(self, spider, outcome, spider_args_override, jobkey):
        pass

    def add_job(self, spider, spider_args_override):
        params = (spider_args_override or {}).copy()
        params["spider"] = spider
        self.__additional_jobs.append(params)

    @abc.abstractmethod
    def set_parameters_gen(self):
        for i in []:
            yield i

    def __add_jobseq_tag(self, params):
        tags = params.setdefault("tags", [])
        jobseq_tag = None
        for tag in tags:
            if tag.startswith("JOBSEQ="):
                jobseq_tag = tag
                break
        if jobseq_tag is None:
            jobseq_tag = f"JOBSEQ={self.__next_job_seq:010d}"
            self.__next_job_seq += 1
        else:
            tags.remove(jobseq_tag)
            jobseq, repn = get_jobseq([jobseq_tag])
            jobseq_tag = f"JOBSEQ={jobseq:010d}.r{repn + 1}"
        tags.append(jobseq_tag)

    def workflow_loop(self):
        self.check_running_jobs()
        while len(self._running_job_keys) < self.max_running_jobs:
            try:
                if self.__additional_jobs:
                    next_params = self.__additional_jobs.pop(0)
                else:
                    next_params = next(self.__parameters_gen)
                spider = next_params.pop("spider", self.spider)
                assert spider is not None, "No spider set."
                spider_args = get_spider_args_from_params(next_params)
                jobid = self.get_job_id({"spider": spider, "spider_args": spider_args})
                if jobid not in self.__jobids:
                    self.__add_jobseq_tag(next_params)
                    self.schedule_spider(spider=spider, spider_args_override=next_params)
                    self.__jobids.add(jobid)
            except StopIteration:
                if self._running_job_keys:
                    break
                return False
        return True

    @staticmethod
    def get_job_id(job):
        jdict = job.get("spider_args", {}).copy()
        jdict["spider"] = job["spider"]
        for k, v in jdict.items():
            jdict[k] = str(v)
        jid = json.dumps(jdict, sort_keys=True)
        return hashstr(jid)

    def resume_workflow(self):
        for job in super().resume_workflow():
            jobid = self.get_job_id(job)
            self.__jobids.add(jobid)
            self.__next_job_seq = max(self.__next_job_seq, get_jobseq(job["tags"])[0] + 1)
        count = 0
        diff_count = 0
        for job in self.get_owned_jobs(spider=self.spider, state=["finished"], meta=["spider_args", "tags", "spider"]):
            jobid = self.get_job_id(job)
            if jobid not in self.__jobids:
                diff_count += 1
                self.__jobids.add(jobid)
            self.__next_job_seq = max(self.__next_job_seq, get_jobseq(job["tags"])[0] + 1)
            count += 1
        _LOG.info(f"Added a total of {count} completed jobs ({diff_count} different ones)")
        _LOG.info(f"Next job sequence number: {self.__next_job_seq}")

    def on_close(self):
        pass
